#!ruby -Ks
# -*- mode: ruby; encoding: sjis -*-
# Last updated: <2014/12/17 02:04:45 +0900>

require 'dxruby'
require 'benchmark'

# BDFtHg`pNX
#
# @version 0.0.3
#
# @note iso8859, jisx0208, iso10646 ݂̂ɑΉ
#
# @example
#    require 'dxruby'
#    require_relative 'bdffont'
#    bdf = BDFFont.new("mplus_j12r_jisx0208.bdf")
#    bdf.add("mplus_f12r_iso8859.bdf")
#    Window.loop do
#      x, y = 10, 10
#      Window.draw_bdffont(x, y, "This is a BDF font", bdf)
#      y += bdf.height
#      Window.draw_bdffont(x, y, "The quick brown fox jumps over the lazy dog.", bdf)
#      y += bdf.height
#      Window.draw_bdffont(x, y, "1234567890 ! = - +", bdf)
#      y += bdf.height
#      Window.draw_bdffont(x, y, "JAPANESE MESSAGE OK", bdf)
#    end
#
class BDFFont

  # @private
  # BDFtHg1̏i[\
  BDFChar = Struct.new(:startchar, :encoding, :image,
                       :ofsx, :ofsy, :w, :h,
                       :swidth, :sheight, :dwidth, :dheight)

  # @private
  # BDFtHg1t@C̃f[^i[NX
  class BDFdata

    # @return [Hash] BDFtHg1pNX𕡐i[nbV
    attr_reader :bdf_dic

    # @return [Integer] R[hB:iso8859, :jisx0208, :iso10646 蓾
    attr_reader :charset

    # @return [Integer] oEfBO{bNX̉
    attr_reader :width

    # @return [Integer] oEfBO{bNX̏c
    attr_reader :height

    # @return [Integer] oEfBO{bNXxItZbg
    attr_reader :ofsx

    # @return [Integer] oEfBO{bNXyItZbg
    attr_reader :ofsy

    # @return [Integer] |CgTCY
    attr_reader :pt_size

    # @return [Integer] DPI
    attr_reader :xdpi

    # @return [Integer] cDPI
    attr_reader :ydpi

    # @return [Integer] sNZTCY
    attr_reader :pixel_size

    # RXgN^
    #
    # @param [String] bdffilename BDFtHgt@C
    # @param [Array] col FBa,r,g,b ׂzŎw
    def initialize(bdffilename, col = [255, 255, 255, 255])
      @bdf_dic = {}
      @pt_size = 0
      @xdpi, @ydpi = 0, 0
      @ofsx, @ofsy = 0, 0
      @width, @height = 0, 0
      @pixel_size = 0
      @charset = nil
      @cahrs = 0

      # t@Cǂݍ
      f = open(bdffilename)
      filedata = f.read
      f.close

      bdfc = nil
      startchar = nil
      charcode = -1
      bitmap_read = false
      bitpat = []

      filedata.each_line do |l|
        # 1s
        l.chomp!

        if startchar
          # 1ǂݎ蒆
          if bitmap_read
            if l.index('ENDCHAR') == 0
              # rbg}bvf[^`IBimage𐶐
              bdfc.image = make_image(bdfc.w, bdfc.h, bitpat, col)
              @bdf_dic[bdfc.encoding] = bdfc
              bitmap_read = false
              startchar = nil
            else
              # rbg}bvz𗭂߂
              bitpat.push(l)
            end
          else
            # 1̃vpeBǂݎ
            p = l.split(' ')
            case p[0]
            when 'BITMAP'
              bitmap_read = true
            when 'ENCODING'
              bdfc.encoding = p[1].to_i
            when 'SWIDTH'
              bdfc.swidth, bdfc.sheight = p[1].to_i, p[2].to_i
            when 'DWIDTH'
              bdfc.dwidth, bdfc.dheight = p[1].to_i, p[2].to_i
            when 'BBX'
              bdfc.w, bdfc.h = p[1].to_i, p[2].to_i
              bdfc.ofsx, bdfc.ofsy = p[3].to_i, p[4].to_i
            end
          end
        else
          # BDF̑S̏ǂݎ
          p = l.split(' ')
          case p[0]
          when 'STARTCHAR'
            # ̃rbg}bvf[^`Jn
            startchar = p[1]
            bdfc = BDFChar.new()
            bdfc.startchar = startchar
            bitmap_read = false
            bitpat.clear
          when 'SIZE'
            # |CgAdpi
            @pt_size, @xdpi, @ydpi = p[1].to_i, p[2].to_i, p[3].to_i
          when 'FONTBOUNDINGBOX'
            # oEfBO{bNX
            @width, @height = p[1].to_i, p[2].to_i
            @ofsx, @ofsy = p[3].to_i, p[4].to_i
          when 'PIXEL_SIZE'
            # sNZTCY
            @pixel_size = p[1].to_i
          when 'CHARSET_REGISTRY'
            # R[h
            @charset = get_charset(p[1])
          when 'CHARS'
            # S
            @chars = p[1].to_i
          when 'ENDFONT'
            # Sf[^ǂݎI
            break
          end
        end
      end

      # puts @bdf_dic.length
    end

    # @private
    # 16i̔z񂩂Image𐶐ĕԂ
    #
    # @param [Integer] w 
    # @param [Integer] h c
    # @param [Array] dt 16i񂪓z
    # @param [Array] col FBa,r,g,b ׂzŎw
    # @return [Object] ImageIuWFNg
    def make_image(w, h, dt, col)
      img = Image.new(w, h)
      msk = 1 << (dt[0].length * 4 - 1)
      v = 0
      h.times do |y|
        v = dt[y].hex
        w.times do |x|
          img[x, y] = col if (v & (msk >> x)) != 0
        end
      end
      return img
    end

    # @private
    # charset̃V{Ԃ
    #
    # @param [String] str charsetނ܂ޕ
    # @return [Integer] charsetV{
    def get_charset(str)
      cst = str.gsub('"', '').downcase
      if cst.index("jisx0208") == 0
        return :jisx0208
      elsif cst.index("jisx0213") == 0
        return :jisx0213
      elsif cst.index("iso10646") == 0
        return :iso10646
      end
      return :iso8859
    end
  end

  # @return [Integer] oEfBO{bNX̏c(ől)
  attr_reader :height

  # @return [Integer] oEfBO{bNXyItZbg
  attr_reader :ofsy

  # RXgN^
  #
  # @param [String] bdffilename BDFtHgt@C
  # @param [Array] col FBa,r,g,b ׂzŎw
  #
  def initialize(bdffilename, col = [255, 255, 255, 255])
    @bdffiles = []
    bdfd = BDFdata.new(bdffilename, col)
    @bdffiles.push(bdfd)
    @width = bdfd.width
    @height = bdfd.height
    @ofsy = bdfd.ofsy
  end

  # BDFt@CǉB
  #
  # BDFtHg(jisx0208)ŃNX𐶐A
  # pBDFtHg(iso8859)ꏏɊ܂߂鎞ɎgB
  #
  # @param [String] bdffilename BDFtHgt@C
  # @param [Array] col FBa,r,g,b ׂzŎw
  #
  def add(bdffilename, col = [255, 255, 255, 255])
    bdfd = BDFdata.new(bdffilename, col)
    @bdffiles.push(bdfd)
    @width = bdfd.width if @width < bdfd.width
    @height = bdfd.height if @height < bdfd.height
    @ofsy = bdfd.ofsy if @ofsy.abs < bdfd.ofsy.abs
  end

  # zƂĎw肵ۂɁABDFtHg1p̃NXԂ
  #
  # @param [Integer] index zY
  # @return [Object] BDFtHg1̃NX
  # @return [nil] nilȂo^͌Ȃ
  #
  def [](index)
    @bdffiles.each do |d|
      code = nil
      case d.charset
      when :jisx0208
        code = self.sjis2jisx0208(index.ord)
      when :jisx0213
        # ΉBϊ@Ȃ
        code = self.sjis2jisx0208(index.ord)
        # code = index.encode("UTF-8", "Windows-31J").ord
        # code = index.encode("EUC-JISX0213", "Windows-31J").ord
        # code = index.encode("EUC-JP-2004", "Windows-31J").ord
      when :iso10646
        code = index.encode("UTF-8", "Windows-31J").ord
      else
        code = index.ord
      end
      return d.bdf_dic[code] if d.bdf_dic.has_key?(code)
    end
    return nil
  end

  # @private
  # SJISR[hJISX0208R[hɕϊĕԂ
  # @param [Integer] code SJISR[h
  # @return [Integer] JISX0208R[h
  def sjis2jisx0208(code)
    code -= 0x4000 if code >= 0x0e000
    ub = (code & 0x0ff00) >> 8
    code = ((ub - 0x081) << (1+8)) + (code & 0x00ff)
    code -= 1 if (code & 0x00ff) >= 0x80
    if (code & 0x00ff) >= 0x009e
      code += (0x0100 - 0x009e)
    else
      code -= 0x0040
    end
    code += 0x2121
    return code
  end

end

# WindowNXɒǉBDFtHg`p\bh
module Window

  # BDFtHg`
  # @param [Number] x `xW
  # @param [Number] y `yW
  # @param [String] str `敶
  # @param [Object] bdf BDFtHgNX
  # @param [Integer] z `D揇
  def self.draw_bdffont(x, y, str, bdf, z = 0)
    str.split('').each do |code|
      b = bdf[code]
      unless b == nil
        x1 = x + b.ofsx
        y1 = y + bdf.height + bdf.ofsy - b.h - b.ofsy
        Window.draw(x1, y1, b.image, z)
        x += b.dwidth
      end
    end
  end
end

# ImageNXɒǉBDFtHg`p\bh
class Image

  # BDFtHg`
  # @param [Number] x `xW
  # @param [Number] y `yW
  # @param [String] str `敶
  # @param [Object] bdf BDFtHgNX
  def draw_bdffont(x, y, str, bdf)
    str.split('').each do |code|
      b = bdf[code]
      unless b == nil
        x1 = x + b.ofsx
        y1 = y + bdf.height + bdf.ofsy - b.h - b.ofsy
        self.draw(x1, y1, b.image)
        x += b.dwidth
      end
    end
  end
end

if $0 == __FILE__
  #eXgR[h

  lst = [
    "This is a BDF font. !\"\#$%&'()*+,-./0123456789:;<=>?@",
    "A B C D EFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~",
    "AVB]̒",
    "e̖SCŏ̎瑹΂BqɓΊpB"
  ]

  col = [255, 128, 255, 128]

  # BDFtHgNX𐶐
  bdfs = []
  Benchmark.bm do |x|
    x.report {
      case 0

      when 0
        bdf = BDFFont.new('bdf/mplus_j10r_jisx0208.bdf', col)
        bdf.add('bdf/mplus_f10r_iso8859.bdf', col)
        bdfs.push(bdf)

        bdf = BDFFont.new('bdf/umplus_j10r_iso10646.bdf', col)
        bdfs.push(bdf)

        bdf = BDFFont.new('bdf/mplus_j12r_jisx0208.bdf', col)
        bdf.add('bdf/mplus_f12r_iso8859.bdf', col)
        bdfs.push(bdf)

      when 1
        bdf = BDFFont.new('bdf/milkjf_k16_jisx0208.bdf', col)
        bdf.add('bdf/milkjf_8x16_iso8859.bdf', col)
        bdfs.push(bdf)

        bdf = BDFFont.new('bdf/shnmk16_jisx0208.bdf', col)
        bdf.add('bdf/shnm8x16a_iso8859.bdf', col)
        bdfs.push(bdf)

      when 2
        bdf = BDFFont.new('bdf/courR12.bdf', col)
        bdfs.push(bdf)

        bdf = BDFFont.new('bdf/courR14.bdf', col)
        bdfs.push(bdf)

      when 3
        # ----------------------------------------
        # Ή
        bdf = BDFFont.new('bdf/warabi12-1.bdf', col)
        bdfs.push(bdf)

        bdf = BDFFont.new('bdf/jiskan16-2004-1_jisx0213.bdf', col)
        bdf.add('bdf/jiskan16-2000-2_jisx0213.bdf', col)
        bdfs.push(bdf)
      end
    }
  end

  # 摜ɕ`悵Ă݂
  image = Image.new(640, 240, [255, 64, 64, 64])
  x, y = 0, 0
  bdfs.each do |bdf|
    lst.each do |s|
      image.draw_bdffont(x, y, s, bdf)
      y += bdf.height
    end
    y += bdf.height
  end

  Window.bgcolor = [96, 96, 96]

  Window.loop do
    break if Input.keyPush?(K_ESCAPE)

    # ʂɕ`悵Ă݂
    x, y = 16, 0
    bdfs.each do |bdf|
      lst.each do |s|
        Window.draw_bdffont(x, y, s, bdf)
        y += bdf.height
      end
      y += bdf.height
    end

    Window.draw(x, 240, image)

    bar = Window.height * Window.getLoad / 100.0
    Window.drawBoxFill(0, 0, 2, bar, C_RED)
  end

end
